/*
 * cmd.c
 * functions for cancan's built-in #commands
 *
 * (created: Finn Arne Gangstad, Dec 25th, 1993)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef STDASCII
#include <ctype.h>
#endif
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#ifdef AIX
#include <sys/select.h>
#endif

#include "cancan.h"
#include "cmd.h"
#include "term.h"
#include "list.h"
#include "beam.h"

int select();
char *strdup();
int setenv();
void unsetenv();
void bzero();
#ifdef NOBZERO
#define bzero(a,b) memset ((a), 0, (b))
#endif

static void show_variables();
static int attr_token();
static int parse_attributes();
static void show_keybinds();
static char *bindtext_by_funct();
static void keybind_modify();
static char *unescape_seq();
static int key_collision();
static char *read_seq();
static int get_one_char();
static void print_seq();
static void show_marks();
static void show_aliases();
static void show_actions();
static int toggle();

static void cmd_action(), cmd_alias(), cmd_bind(), cmd_cancel();
static void cmd_capture(), cmd_echo(), cmd_edit(), cmd_for();
static void cmd_help(), cmd_hilite(), cmd_identify(), cmd_init();
static void cmd_map(), cmd_mark(), cmd_prefix(), cmd_quote();
static void cmd_read(), cmd_record(), cmd_retrace(), cmd_set();
static void cmd_speedwalk(), cmd_ver(), cmd_wrap();
#ifndef OMIT_SESSIONS
static void cmd_connect(), cmd_snoop(), cmd_zap();
#endif
#ifndef NOSHELL
static void cmd_shell();
#endif
#ifdef TELNETBUG
static void cmd_stdcolour();
#endif

cmdstruct commands[] = 
{
   {"help", "[topic]\t\t\thelp on commands, or this text", cmd_help},
   {"17", "command\t\t\trepeat 'command' 17 times", (void(*)())NULL},
#ifndef NOSHELL
   {"!", "shell-command\t\texecute a shell command using /bin/sh", cmd_shell},
   {"?", "shell-command;...[;...]\tif .. then .. else", 0},
#endif
   {"action", "[name][\"pattern\"][=cmd]\tlist/delete/define actions",
	cmd_action},
   {"alias", "[name[=text]]\t\tlist/delete/define aliases", cmd_alias},
   {"bind", "[name[=command]]\t\tlist/delete/define key bindings", cmd_bind},
   {"cancel", "[number]\t\tcancel editing session", cmd_cancel},
   {"capture", "[filename]\t\tbegin/end of capture to file", cmd_capture},
#ifndef OMIT_SESSIONS
   {"connect", "[id[ init][=address]]\tconnect a new session",
	cmd_connect},
#endif
   {"echo", "[-n ][text]\t\tdisplay text on screen", cmd_echo},
   {"edit", "\t\t\t\tlist editing sessions", cmd_edit},
   {"for", "a b [c] cmd|var cmd\titerate a command", cmd_for},
   {"hilite", "[attr]\t\t\thighlight your input line", cmd_hilite},
   {"identify", "[startact endact]\tsend MUME client identification",
	cmd_identify},
   {"init", "[-][[+]string]\t\tsend string to host on connect", cmd_init},
   {"map", "[[-]number]\t\tshow/clear (auto)map", cmd_map},
   {"mark", "[string[=attr]]\t\tlist/define markers", cmd_mark},
   {"prefix", "string\t\t\tprefix all lines with string", cmd_prefix},
   {"quote", "[on|off]\t\t\ttoggle verbatim-flag", cmd_quote},
   {"read", "file\t\t\tread lines from file/pipe and send", cmd_read},
   {"record", "[name]\t\t\trecord typed lines into alias", cmd_record},
   {"retrace", "[[-]number]\t\tretrace the last number steps", cmd_retrace},
   {"set", "name[=value]\t\tset variable value", cmd_set},
#ifndef OMIT_SESSIONS
   {"snoop", "id\t\t\ttoggle output display for session", cmd_snoop},
#endif
   {"speedwalk", "[on|off]\t\ttoggle speedwalk mode", cmd_speedwalk},
#ifdef TELNETBUG
   {"stdcolour", "attr\t\t\tset standard background colour", cmd_stdcolour},
#endif
   {"ver", "\t\t\t\tshow cancan version", cmd_ver},
   {"wrap", "[on|off]\t\t\tturn word wrapping on/off", cmd_wrap},
#ifndef OMIT_SESSIONS
   {"zap", "id\t\t\t\tdisconnect a session", cmd_zap},
#endif
   {(char*)NULL, (char*)NULL, (void(*)())NULL}
};

/* anyone knows if ANSI 6429 talks about more than 8 colours? */
char *colournames[] = {
    "black", "red", "green", "yellow", "blue", "magenta", "cyan", "white"
};

char *recording_name;		/* name of recording alias or NULL */

static void cmd_record(arg)
char *arg;
{
    if(*arg) {
	if(recording_name) {
	    printf("#already recording to alias '%s'\n", recording_name);
	    free(recording_name);
	}
	strtok(arg, " ");	/* strip any accidental trailing blanks */
	recording_name = strdup(arg);
	if(verbose) printf("#recording into alias '%s'\n", recording_name);
    } else if(recording_name) {
	if(verbose) printf("#recording of alias '%s' stopped.\n", recording_name);
	free(recording_name);
	recording_name = NULL;
	new_settings = 1;
    } else
	printf("#not recording any alias\n");
}

static void cmd_echo(arg)
char *arg;
{
    char *pat = "%s\n";
    if(!strncmp(arg, "-n", 2)) {
	if(!arg[2]) return;
	if(arg[2] == ' ') {
	    pat = "%s";
	    arg += 3;
	}
    }
    printf(pat, unescape_seq(arg));
}

static void show_variables()
{
    extern char **environ;
    int i;
    for(i = 0; environ[i]; i++)
	printf("%s\n", environ[i]);
}

static void cmd_set(arg)
char *arg;
{
    char *s;
    if(!*arg)
	show_variables();
    else if((s = strchr(arg, '=')) && (*s = 0, !*++s))
	unsetenv(arg);
    else
	setenv(arg, s ? s : "", 1);
}        

static void cmd_init(arg)
char *arg;
{
  if (*arg == '-') 
   {
     initstr[0] = 0;
     printf("#init cleared.\n");
   }
  else 
   {
     if (*arg == '+')
     	strcat(initstr, arg+1);
     else if (*arg)
     	strcpy(initstr, arg);
     printf("#init: %s\n", initstr);
   }
  new_settings = 1;
}

static void cmd_hilite(arg)
char *arg;
{
    int attr = parse_attributes(arg);
    if(attr >= 0) {
	attr_string(attr, edattrbeg, edattrend);
	printf("#input highlighting is now %so%s%s.\n", 
	       edattrbeg, (attr == NOATTRCODE) ? "ff" : "n", edattrend);
    }
}

/*
 * return token number for attribute word or
 * -1 for unrecognized word.
 */
static int attr_token(word)
char *word;
{
    int i;
    for(i = 0; i < COLOURS; i++)
	if(!strcmp(word, colournames[i])) return i;
    if(!strcmp(word, "bold"))
	return BOLD;
    else if(!strcmp(word, "underline"))
	return UNDERLINE;
    else if(!strcmp(word, "inverse") || !strcmp(word, "reverse"))
	return INVERSE;
    else if(!strcmp(word, "on"))
	return NOATTRIB;	/* fill word */
    return -1;
}

/*
 * parse attribute description in line.
 * Return attribute if successful, -1 otherwise.
 */
static int parse_attributes(line)
char *line;
{
    int t, bg;
    char *p = strtok(line, " ");
    if(!p) return NOATTRCODE;
    for( ; ; ) {
	if((t = attr_token(p)) < 0) break;
	p = strtok(NULL, " ");
	if(t > COLOURS) {	/* bold/inverse/underline */
	    if(p) break;
	    return ATTRCODE(t, NOATTRIB);
	}
	if(t != NOATTRIB) {
	    if(!p) return ATTRCODE(t, NOATTRIB);
	    if(attr_token(p) != NOATTRIB || !(p = strtok(NULL, " "))) break;
	}
	if((bg = attr_token(p)) < 0 || bg >= COLOURS) break;
	return ATTRCODE(t, bg);
    }
    printf("#invalid attribute syntax\n");
    printf("Valid options are:\tbold, underline, inverse\n");
    printf("\tor:\t\t[foreground] [on background]\nColours:");
    for(t = 0; t < COLOURS; t++)
	printf(" %s%c", colournames[t], (t == COLOURS - 1) ? '\n' : ',');
    return -1;
}

/*
 * return a static pointer to name of given attribute code
 */
char *attr_name(attrcode)
int attrcode;
{
    static char name[32];	/* should be enough */
    unsigned fg = FOREGROUND(attrcode), bg = BACKGROUND(attrcode);

    if(bg == NOATTRIB) {
	if(fg < COLOURS) return colournames[fg];
	switch(fg) {
	    case BOLD: return "bold";
	    case UNDERLINE: return "underline";
	    case INVERSE: return "inverse";
	}
    }
    if(fg < COLOURS)
	sprintf(name, "%s on %s", colournames[fg], colournames[bg]);
    else
	sprintf(name, "on %s", colournames[bg]);
    return name;
}

/*
 * put escape sequences to turn on/off an attribute in given buffers
 */
void attr_string(attrcode, begin, end)
int attrcode; char *begin, *end;
{	
    int fg = FOREGROUND(attrcode), bg = BACKGROUND(attrcode);
    *begin = *end = '\0';
    if(fg < COLOURS) {
	if(bg < COLOURS)
	    sprintf(begin, "\033[3%d;4%dm", fg, bg);
	else
	    sprintf(begin, "\033[3%dm", fg);
    } else if(bg < COLOURS) {
	sprintf(begin, "\033[4%dm", bg);
    } else if(fg == BOLD) {
	if(modebold[0])
	    strcpy(begin, modebold);
	else {
	    strcpy(begin, modestandon);
	    strcpy(end, modestandoff);
	}
    } else if(fg == UNDERLINE) {
	if(modeuline[0])
	    strcpy(begin, modeuline);
	else {
	    strcpy(begin, modestandon);
	    strcpy(end, modestandoff);
	}
    } else if(fg == INVERSE) {
	if(modeinv[0])
	    strcpy(begin, modeinv);
	else {
	    strcpy(begin, modestandon);
	    strcpy(end, modestandoff);
	}
    }
    if(attrcode != NOATTRCODE && !*end)
	strcpy(end, modenorm);
}

static void cmd_mark(arg)
char *arg;
{
    if(!*arg)
	show_marks();
    else
	parse_marker(arg);
}

static void cmd_bind(arg)
char *arg;
{
    if(!*arg)
	show_keybinds();
    else
	parse_bind(arg);
}

/*
 * list keyboard bindings
 */
static void show_keybinds()
{
    keynode *p;
    int count = 0;
    for(p = keydefs; p; p = p->next) {
	if(p->call_data) {
	    if(count == 0)
		printf("#user-defined keys:\n");
	    printf("#bind %s=%s\n", p->name, p->call_data);
	    count++;
	}
    }
    if(count == 0)
	printf("#No key bindings defined right now.\n");
}

/*
 * parse the #bind command 
 */
void parse_bind(str)
char *str;
{
    keynode **np;
    char *text;
    char *rest = chop_arg(str, '=');
    char *kseq = chop_arg(str, ' ');

    if(kseq && !*(kseq = unescape_seq(skipspace(kseq)))) kseq = NULL;
    np = lookup_key(str);
    if(*np) {
	if(!(text = (*np)->call_data) &&
		!(text = bindtext_by_funct((*np)->funct))) return;
	if(kseq && strcmp(kseq, (*np)->sequence))
	    printf("#key '%s' already bound; will not rebind.\n", str);
	else if(!rest) {
	    char buf[LINELEN];
	    sprintf(buf, "#bind %s=%s", str, text);
	    escape_command(buf, inserted_next);
	} else if(!*rest) {
	    delete_keynode(np);
	    if(verbose)
		printf("#key binding deleted: %s=%s\n", str, text);
	    new_settings = 1;
	} else if(strcmp(rest, text)) {
	    if((*np)->call_data)
		free((*np)->call_data);
	    (*np)->funct = key_run_command;
	    keybind_modify(*np, rest);
	    if(verbose)
		printf("#redefined key: %s=%s\n", str, rest);
	    new_settings = 1;
	}
    } else if(!rest || !*rest) {
	printf("#no such key: '%s'\n", str);
    } else if(kseq ? !key_collision(kseq, 1) :
	    !key_collision(kseq = read_seq(), 0)) {
	np = add_keynode(str, kseq, key_run_command, NULL);
	keybind_modify(*np, rest);
	if(verbose)
	    printf("#new key binding: %s=%s\n", str, rest);
	new_settings = 1;
    }
}

/*
 * given a pointer to edit function, return the name of the function:
 * e.g. if(f == prev_char) ==> return "&prev-char"
 */
static char *bindtext_by_funct(f)
void (*f)();
{
    struct defbind_node *p = defbinds;
    while(p->funct && p->funct != f) p++;
    if(!p->name)
	printf("#key '%s' is predefined and cannot be modified.\n", p->label);
    return p->name;
}

/*
 * assign keybind's call_data and change funct if necessary.
 * if keybind is found to be a default one, null call_data.
 */
static void keybind_modify(np, text)
keynode *np;
char *text;
{
    struct defbind_node *p;
    if(*text == '&')
	for(p = defbinds; p->label[0]; p++)
	    if(p->name && !strcmp(p->name, text)) {
		np->funct = p->funct;
		if(p->seq && !strcmp(np->sequence, p->seq) &&
			!strcmp(np->name, p->label)) {
		    np->call_data = NULL;  return;
		}
	    }
    np->call_data = strdup(text);
}

/*
 * unescape, but also recognize ^c syntax.
 */
static char *unescape_seq(str)
char *str;
{
    char *s = str, *d = str;
    do switch(*s) {
	case '^': *(s + 1) = CONTROL(*(s + 1));
	/* FALLTHROUGH */
	case '\\': s++;
    } while((*d++ = *s++));
    return str;
}

/*
 * make an escape sequence printable (for use in definition-files)
 */
void escape_seq(seq, buf)
char *seq, *buf;
{
    char *p = buf;
    /*
     * rules: escape =\^ with \. control chars are written as ^X.
     */
    for(;*seq; seq++) {
	if(*seq < ' ') {
	    *p++ = '^';
	    *p++ = *seq | '@';
	    continue;
	}
	if(*seq == '=' || *seq == '\\' || *seq == '^')
	    *p++ = '\\';
	*p++ = *seq;
    }
    *p = '\0';
}

static int key_collision(keyseq, shadow)
char *keyseq;
int shadow;
{
    keynode **np;
    if(!*keyseq) return 1;
    if(isprint(*keyseq)) {
	printf("#that is not a redefineable key.\n");
	return 1;
    }
    for(np = &keydefs; *np; np = &(*np)->next) {
	register char *a = (*np)->sequence;
	register char *b = keyseq;
	do if(!*a || !*b) {
	    if(*a || *b) {
		printf("#colliding keysequence: key '%s'\n", (*np)->name);
	    } else if((a = (*np)->call_data) ||
		    ((a = bindtext_by_funct((*np)->funct)) && !shadow)) {
		printf("#key already bound as: %s=%s\n", (*np)->name, a);
	    } else if(a) {
		delete_keynode(np);  return 0;
	    }
	    return 1;
	} while(*a++ == *b++);
    }
    return 0;
}

/*
 * prompt the user to press a key, read a single escape sequence from the
 * keyboard, print it. return static pointer to sequence.
 */
static char *read_seq()
{
    static char kbuf[CAPLEN];
    int i = 0, delay = 0;
    printf("#please press the key you want to define: ");  fflush(stdout);
    while(i < CAPLEN - 1 && (kbuf[i] = get_one_char(delay)))
	if(i++ || kbuf[0] != '\033') delay = KBD_TIMEOUT;
    kbuf[i] = 0;
    print_seq(kbuf);  newline();
    return kbuf;
}

/*
 * read a single character from stdin, with timeout in microseconds.
 * timeout == 0 means wait indefinitely (no timeout).
 * return char or 0 if timeout was reached.
 */
static int get_one_char(timeout)
int timeout;
{
    extern int errno;
    int err;
    fd_set fds;
    unsigned char c = 0;
    struct timeval timeoutbuf;

    FD_ZERO(&fds);
    FD_SET(0, &fds);
    timeoutbuf.tv_sec = 0;
    timeoutbuf.tv_usec = timeout;
    while((err = select(2, &fds, NULL, NULL, timeout ? &timeoutbuf : NULL)) < 0)
	if(errno != EINTR) { syserr("select");  break; }
    if(err && read(0, &c, 1) != 1)
	syserr("read from stdin");
    return c;
}

/*
 * print an escape sequence human-readably
 */
static void print_seq(seq)
char *seq;
{
    while(*seq) {
	if(*seq == '\033')
	    printf("esc ");
	else if(*seq == ' ')
	    printf("space ");
	else if(*seq < ' ')
	    printf("^%c ", *seq | '@');
	else
	    printf("%c ", *seq);
	seq++;
    }
}

static void cmd_speedwalk(arg)
char *arg;
{
    if(toggle(&speedwalk, arg) || !parse_walk(arg))
	printf("#speedwalk is now o%s.\n", speedwalk ? "n" : "ff");
}

static void cmd_map(arg)
char *arg;
{
    if (!*arg || *skipnum(arg))
	show_automap();
    else		/* remove some steps without walking */
	remove_steps(atoi(arg));
}

static void cmd_retrace(arg)
char *arg;
{
    retrace_steps(atoi(arg));
}

static void cmd_alias(arg)
char *arg;
{
    if(!*arg)
	show_aliases();
    else
	parse_alias(arg);
}

static void cmd_action(arg)
char *arg;
{
    if(!*arg)
	show_actions();
    else
	parse_action(arg);
}

static void cmd_capture(arg)
char *arg;
{
    if(capfile) {
	if(*arg)
	    printf("#capture already active.\n");
	else {
	    fclose(capfile);
	    capfile = NULL;
	    printf("#end of capture to file.\n");
	}
    } else {
	char *mode = (*arg == '>') ? (arg++, "a") : "w";
	if(!*arg)
	    printf("#capture to what file?\n");
	else if(!(capfile = fopen(arg, mode)))
	    printf("#error writing file %s\n", arg);
	else
	    printf("#capture to %s active, '#capture' ends.\n", arg);
    }
}

static void cmd_prefix(arg)
char *arg;
{
    strcpy(prefixstr, arg);
    if(verbose) printf("#prefix %s.\n", *arg ? "set" : "cleared");
}

#ifndef NOSHELL
static void cmd_shell(arg)
char *arg;
{
    if(*arg) {
	reset_terminal();
	execute(arg);
	set_terminal();
	gotoxy(0, line0 = lines_1);
	fputs(clreoln, stdout);
    }
}
#endif

static void cmd_quote(arg)
char *arg;
{
    toggle(&verbatim, arg);
    printf("#%s mode.\n", verbatim ? "verbatim" : "normal");
}

static void cmd_read(arg)
char *arg;
{
    FILE *fp;
    char buf[LINELEN];
    if(*arg) {
	fp = (arg[0] == '|') ? popen(arg + 1, "r") : fopen(arg, "r");
	if(fp) {
	    while(fgets(buf, LINELEN, fp))
		send_to_host(buf, strlen(buf));
	    if(arg[0] == '|') pclose(fp); else fclose(fp);
	} else
	    printf("#error opening %s\n", arg);
    } else
	printf("#syntax: #read filename or #read |command\n");
}

static void cmd_identify(arg)
char *arg;
{
    edit_start[0] = edit_end[0] = 0;
    if(*arg) {
	char *p = chop_arg(arg, ' ');
	strcpy(edit_start, arg);
	if(p) strcpy(edit_end, p);
    }
    if(!mp_active) {
	mp_active = 1;
	send_to_host(MCIDENT, strlen(MCIDENT));
	data_sent--;		/* MUME doesn't prompt ident string */
    }
}

/*ARGSUSED*/
static void cmd_edit(arg)
char *arg;
{
    editsess *s;
    if(edit_sess) {
	for(s = edit_sess; s; s = s->next)
	    printf("#%s session %d \"%.*s\"\n",
			s->mode == 'E' ? "editing" :
			s->mode == 'V' ? "viewing" :
			s->mode == 'C' ? "cancelled" : "",
			s->num, s->dlen, s->desc);
    } else
	printf("#no active editors.\n");
}

static void cmd_cancel(arg)
char *arg;
{
    editsess *sp;
    int num;
    if(!edit_sess) {
	printf("#no active sessions to cancel.\n");
    } else if(*arg) {
	if((num = atoi(arg))) {
	    for(sp = edit_sess; sp && num != sp->num; sp = sp->next);
	    if(sp) {
		cancel_edit(sp);
		return;
	    }
	}
	printf("#no such session - use the session number\n");
    } else if(edit_sess->next) {
	printf("#several sessions active, use #cancel <number>\n");
    } else
	cancel_edit(edit_sess);
}

/*ARGSUSED*/
static void cmd_help(arg)
char *arg;
{
    int i;
#ifdef HELPFILE
    if(*arg) {
	char line[LINELEN];
	FILE *f = fopen(HELPFILE, "r");
	if(!f)
	    printf("#can't open help file %s\n", HELPFILE);
	else {
	    if(*arg == '#') arg++;
	    while(fgets(line, LINELEN, f))
		if(line[0] != '@' || (line[1] > '\n' &&
			strncmp(line + 1, arg, strlen(arg)))) continue;
		else {
		    while(fgets(line, LINELEN, f) && line[0] != '@')
			fputs(line, stdout);
		    return;
		}
	    printf("#no entry for '%s' in the help file.\n", arg);
	}
    } else
#endif
    {
	printf("# commands availible:\n");
	for(i = 0; commands[i].name; i++)
	    printf("#%s %s\n", commands[i].name, commands[i].help);
    }
}

static void cmd_for(arg)
char *arg;
{
    if(!isdigit(*arg)) {
	char *var, *cmd = chop_arg(arg, ' ');
	int v, last = -1;
	if(cmd) {
	    while((var = getenv(arg)) && (v = atoi(var)) &&
		    (v < 0 ? v = -v : v) < (unsigned)last) {
		last = v;
		parse(cmd, 0, NULL);
	    }
	    return;
	}
    } else {
	static char buf[32], *s;
	int i, to, step;
	if((i = strtol(arg, &s, 10), *s == ' ') &&
		(to = strtol(s + 1, &s, 10), !*s || *s == ' ')) {
	    if(!*s || !(step = strtol(++s, &arg, 10)) || *arg != ' ')
		step = 1;
	    else
		s = arg + 1;
	    while((step > 0) ? (i <= to) : (i >= to)) {
		sprintf(buf, "%d", i);
		setenv("I", buf, 1);
		parse(s, 0, NULL);
		i += step;
	    }
	    unsetenv("I");
	    return;
	}
    }
    printf("#for: bad syntax\n");
}

#ifdef TELNETBUG
static void cmd_stdcolour(arg)
char *arg;
{
    int attrcode = parse_attributes(arg);
    if(attr >= 0) {
	int bg = BACKGROUND(attrcode), fg = FOREGROUND(attrcode);
	if(fg >= COLOURS || bg >= COLOURS) {
	    printf("#please specify foreground and background colours\n");
	} else {
	    sprintf(modenorm, "\033[;3%d;4%dm", fg, bg);
	    printf("%s#standard colour set.\n", modenorm);
	    if(edattrend[0]) strcpy(edattrend, modenorm);
	}
    }
}
#endif

static void cmd_wrap(arg)
char *arg;
{
    toggle(&wrap, arg);
    printf("#word wrap %sabled.\n", wrap ? "en" : "dis");
}

/*ARGSUSED*/
static void cmd_ver(arg)
char *arg;
{
    printver();
}

#ifndef OMIT_SESSIONS
static void cmd_connect (arg)
char *arg;
{
    if(!*arg)
	show_connections();
    else {
	char *host = chop_arg(arg, '=');
	char *init = chop_arg(arg, ' ');
	if(find_sess(arg, 0)) {
	    printf("#session with that name already exists.\n");
	} else if(host) {
	    char *port = chop_arg(host, ' ');
	    connect_sess(arg, init, host, port ? atoi(port) : 0);
	} else
	    connect_sess(arg, init, NULL, 0);
    }
}

static void cmd_zap (arg)
char *arg;
{
  if (!*arg)
     printf("#zap: no session name\n");
  else
     disconnect_sess(arg);
}

static void cmd_snoop (arg)
char *arg;
{
  if (!*arg)
     printf("#snoop: no session name\n");
  else
     snoop_sess(arg);
}
#endif

/*
 * parse the #alias command
 */
void parse_alias(str)
char *str;
{
    aliasnode **np;
    char *rest = chop_arg(str, '=');
    if(strchr(str, ' ')) {
	printf("#alias name should not contain a space.\n");
	return;
    }
    np = lookup_alias(str);
    if(*np) {
	if(!rest) {
	    char buf[LINELEN];
	    sprintf(buf, "#alias %s=%s", str, (*np)->subst);
	    escape_command(buf, inserted_next);
	} else if(!*rest) {
	    if(verbose)
		printf("#deleting alias: %s=%s\n", str, (*np)->subst);
	    delete_aliasnode(np);
	    new_settings = 1;
	} else if(strcmp(rest, (*np)->subst)) {
	    free((*np)->subst);
	    (*np)->subst = strdup(rest);
	    if(verbose)
		printf("#changed alias: %s=%s\n", str, rest);
	    new_settings = 1;
	}
    } else if(rest && *rest) {
	if(strcmp(str, "#alias")) {
	    add_aliasnode(str, rest);
	    if(verbose)
		printf("#new alias: %s=%s\n", str, rest);
	    new_settings = 1;
	} else
	    printf("#won't alias \"#alias\"\n");
    } else
	printf("#unknown alias: %s\n", str);
}

/*
 * parse the #action command
 */
void parse_action(str)
char *str;
{
    actionnode **np = 0;
    char *label = 0, *pattern = 0, *rest = 0;
    char *s = str, sign = 0;
    if(*s != '"' && (
	    (s = chop_arg(
		    label = (*s != '-' && *s != '+' ? s : (sign = *s, s + 1)),
		    ' ') ),
	    (rest = chop_arg(label, '=')) )) {
	if(s) *(s - 1) = ' ';
    } else if(s) {
	if(*s == '"')
	    if(!(s = chop_arg(pattern = s + 1, '"'))) {
		printf("#action: pattern not \" terminated\n");  return;
	    } else
		unescape(pattern);
	if(*(s = skipspace(s)))
	    if(*s != '=') {
		printf("#action: expecting a =\n");  return;
	    } else
		rest = s + 1;
    }
    /* at this point, label and/or pattern are set */
    np = label ? lookup_action(label) : lookup_action_pattern(pattern);
    if(*np) {
	if(!rest) {
	    if(sign) {
		(*np)->active = sign == '+';
		if(verbose)
		    printf("#action %s \"%s\" is now o%s\n", (*np)->label,
			    (*np)->pattern, (sign == '+') ? "n" : "ff");
	    } else {
		char buf[LINELEN];
		sprintf(buf, "#action %s%s \"",
			(*np)->active ? "" : "-", (*np)->label);
		rest = buf + strlen(buf);
		escape_specials((*np)->pattern, buf + strlen(buf), "\\\";");
		sprintf(buf + strlen(buf), "\"=%s", (*np)->commands);
		escape_command(buf, inserted_next);
	    }
	} else if(!*rest) {
	    if(verbose)
		printf("#action %s \"%s\" deleted\n",
			(*np)->label, (*np)->pattern);
	    delete_actionnode(np);
	    new_settings = 1;
	} else {
	    if(sign && (sign = (*np)->active != (sign == '+')))
		(*np)->active ^= 1;
	    if(pattern && strcmp((*np)->pattern, pattern)) {
		free((*np)->pattern);
		(*np)->pattern = strdup(pattern);
		sign = 1;
	    }
	    if(strcmp((*np)->commands, rest)) {
		free((*np)->commands);
		(*np)->commands = strdup(rest);
		sign = 1;
	    }
	    if(sign) {
		if(verbose) printf("#changed action: %s%s \"%s\"=%s\n",
			(*np)->active ? "" : "-", (*np)->label,
			(*np)->pattern, (*np)->commands);
		new_settings = 1;
	    }
	}	    	    
    } else if(!pattern) {
	printf("#no action labelled '%s'\n", label);
    } else if(!rest || !*rest) {
	printf("#no action with pattern \"%s\"\n", pattern);
    } else {
	static last = 0;
	static char lbuf[16];
	if(!label) {
	    while(sprintf(lbuf, "%d", ++last), *lookup_action(lbuf));
	    label = lbuf;
	}
	add_actionnode(pattern, rest, label, sign != '-');
	if(verbose)
	    printf("#new action: %s \"%s\"=%s\n", label, pattern, rest);
	new_settings = 1;
    }
}

/*
 * parse arguments to the #mark command
 */
void parse_marker(str)
char *str;
{
    int attrcode;
    marknode **np;
    char *rest = chop_arg(str, '=');
    unescape(str);
    np = lookup_marker(str);
    if(*np) {
	if(!rest) {
	    strcpy(inserted_next, "#mark ");
	    rest = inserted_next + 6;
	    if(*str == ' ') *rest++ = '\\';
	    escape_specials(str, rest, "\\=;");
	    sprintf(rest + strlen(rest), "=%s", attr_name((*np)->attrcode));
	} else if(!*rest) {
	    if(verbose)
		printf("#mark deleted: %s=%s\n",
			(*np)->pattern, attr_name((*np)->attrcode));
	    delete_marknode(np);
	    new_settings = 1;
	} else if((attrcode = parse_attributes(rest)) >= 0 &&
		(*np)->attrcode != attrcode) {
	    (*np)->attrcode = attrcode;
	    if(verbose)
		printf("#changed mark: %s=%s\n",
			str, attr_name((*np)->attrcode));
	    new_settings = 1;
	}
    } else if(!rest || !*rest) {
	printf("#unknown marker: '%s'\n", str);
    } else if((attrcode = parse_attributes(rest)) >= 0) {
	np = add_marknode(str, attrcode);
	if(verbose)
	    printf("#new mark: %s=%s\n", str, attr_name((*np)->attrcode));
	new_settings = 1;
    }
}

/*
 * show defined marks
 */
static void show_marks()
{
    marknode *p;
    printf("#%s marker%s defined%c\n", markers ? "The following" : "No",
	   (markers && !markers->next) ? " is" : "s are",
	   markers ? ':' : '.');
    for(p = markers; p; p = p->next)
	printf("#mark %s=%s\n", p->pattern, attr_name(p->attrcode));
}

/*
 * show defined aliases
 */
static void show_aliases()
{
    aliasnode *p;
  
    printf("#%s alias%s defined%c\n", aliases ? "The following" : "No",
	   (aliases && !aliases->next) ? " is" : "es are",
	   aliases ? ':' : '.');
    for(p = aliases; p; p = p->next)
	printf("#alias %s=%s\n", p->name, p->subst);
}

/*
 * show defined actions
 */
static void show_actions()
{
    actionnode *p;
  
    printf("#%s action%s defined%c\n", actions ? "The following" : "No",
	   (actions && !actions->next) ? " is" : "s are", actions ? ':' : '.');
    for(p = actions; p; p = p->next)
	printf("#action %c%s \"%s\"=%s\n", p->active ? '+' : '-', p->label,
	       p->pattern, p->commands);
}

/*
 * if arg is either "on" or "off", toggle a variable and return 1.
 * otherwise return 0.
 */
static int toggle(var, arg)
int *var;
char *arg;
{
    if(!strcmp(arg, "on"))
	*var = 1;
    else if(!strcmp(arg, "off"))
	*var = 0;
    else
	return 0;
    return 1;
}
